/****************************************************************************/
/*! \file
**  \brief  Versions (wfx_Versions) file operations (including HTTP)
**
**  This file implements the file operations (save/load) and HTTP transfer.
**
**  \n
**  Copyright  ©  aNDreas Bolotă  aNDreas·Bolota(α)kirnbie·com
**
**  Versions (wfx_Versions) allows you to find out the current version number
**  of any program downloadable online.
**  Versions (wfx_Versions) emulates the behavior of Versions © Fabio Chelly
**  file-system plugin for Total Commander © Christian Ghisler.
**
**  This program is free software: you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation, either version 3 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program.  If not, see <http://www.gnu.org/licenses/>
*/


/****************************************************************************/
/*! \brief  Holds the current HTTP Request object
**
**  Replaced by next .list item after current request status is "complete"
*/
var  wfxVersions_dialogVars_httpRequest = null;


/****************************************************************************/
/*! \brief  Validates if a string is a valid URL
**
**  \param[in]  url  the candidate URL to be validated
**
**  \return  boolean (true = valid / false = invalid)
*/
function  wfxVersions_httpOps_isValidURL(  url  )
{
  try
  {
    var  oFix = Components.classes[ "@mozilla.org/docshell/urifixup;1" ]
                          .createInstance( Components.interfaces.nsIURIFixup )
                          .createFixupURI(  url.replace( " ", "%20" ),
                                            Components.interfaces.nsIURIFixup.FIXUP_FLAGS_MAKE_ALTERNATE_URI  );

    /*  FIXUP_FLAGS_MAKE_ALTERNATE_URI may have added a "www." in front
    **  but the URL is valid also without it
    */
    if((  url != oFix.spec  )&&(  url != oFix.spec.replace( /www\./i, "" )  ))
    {
      return  false;
    }
  }
  catch(  e  )
  {
    /*  a really bad url-string causes createFixupURI() to crash  */
    return  false;
  }

  return  true;
}


/****************************************************************************/
/*! \brief  Workaround for a problem in Firefox 3.1 beta builds
**
**  HTTP requests trigger on-error events after having been aborted,
**  so we an-null all on-event notifications prior to the abort
**
**  \param  --
**
**  \return  --
*/
function  wfxVersions_httpOps_ghost()
{
  var  oldReq = wfxVersions_dialogVars_httpRequest;
  wfxVersions_dialogVars_httpRequest = null;

  oldReq.onerror    = null;
  oldReq.onload     = null;
  oldReq.onprogress = null;

  oldReq.abort();
}


/****************************************************************************/
/*! \brief  Abort a HTTP Request in progress
**
**  Caused by a timeout after launching the request
**
**  \param[in]  iItem   element scanned
**  \param[in]  bForce  force abort regardless of timeout
**
**  \return  --
*/
function  wfxVersions_httpOps_abort(  iItem,  bForce  )
{
  if((  wfxVersions_dialogVars_httpRequest == null  )||(  wfxVersions_dialogVars_itemsList == null  ))
  {
    return;
  }

  if((  iItem != wfxVersions_dialogVars_progress  )||(  wfxVersions_dialogVars_itemsList[iItem].status != 2  ))
  {
    return;
  }

  /*  retrieve the wfxVersions HTTP Request timeout duration  */
  var  iDelay = -1;
  if(  !bForce  )
  {
    iDelay = Components.classes[ "@mozilla.org/preferences-service;1" ]
                       .getService(  Components.interfaces.nsIPrefBranch  )
                       .getIntPref(  "extensions.wfx_Versions.httpReqTimeout"  );
  }

  if((  iDelay > wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].step  )||(  iDelay == 0  ))
  {
    /*  timeout not expired, or timeout disabled  */
    wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].step++;
    wfxVersions_globalOps_setTimeout(  "wfxVersions_httpOps_abort('" + iItem + "',false)",  999  );
  }
  else
  {
    /*  offer some more details for current error  */
    wfxVersions_dialogOps_notifyOnError(  wfxVersions_dialogOps_localeParams( "szTimeout" ),  wfxVersions_dialogOps_localeParams( "szHTimeout" )  );
  }
}


/****************************************************************************/
/*! \brief  Handles errors during HTTP Requests
**
**  For each URL there are two requests, once for HEAD and once for GET.
**  Exceptions are handled by same function (this function)
**
**  \param[in]  eHTTP  the caught exception
**
**  \return  --
*/
function  wfxVersions_httpOps_onError(  eHTTP  )
{
  if((  wfxVersions_dialogVars_httpRequest == null  )||(  wfxVersions_dialogVars_itemsList == null  ))
  {
    return;
  }

  var  xOld =
  {
    szEditHome:  document.getElementById(  "wfxVersions_dialogIDs_editHome"  ).value,
    szEditDown:  document.getElementById(  "wfxVersions_dialogIDs_editDown"  ).value,
    szEditURL:   document.getElementById(  "wfxVersions_dialogIDs_editURL"  ).value,
    szVersion:   wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].version,
    szScanHome:  wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].home,
    szScanDown:  wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].down,
    szScanURL:   wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].url
  };

  document.getElementById(  "wfxVersions_dialogIDs_editHome"  ).value    = document.getElementById(  "wfxVersions_dialogIDs_editHome"  ).previousSibling.getAttribute(  "tooltiptext"  );
  document.getElementById(  "wfxVersions_dialogIDs_editDown"  ).value    = document.getElementById(  "wfxVersions_dialogIDs_editDown"  ).previousSibling.getAttribute(  "tooltiptext"  );
  document.getElementById(  "wfxVersions_dialogIDs_editURL"  ).value     = document.getElementById(  "wfxVersions_dialogIDs_editURL"   ).previousSibling.getAttribute(  "tooltiptext"  );
  wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].home = wfxVersions_dialogOps_likeMRT(  xOld.szScanHome,  xOld.szVersion  );
  wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].down = wfxVersions_dialogOps_likeMRT(  xOld.szScanDown,  xOld.szVersion  );
  wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].url  = wfxVersions_dialogOps_likeMRT(  xOld.szScanURL,   xOld.szVersion  );

  var  szTxt = wfxVersions_dialogOps_localeParams(  "szException"  ).replace(  "[E]",  eHTTP.name  );
  wfxVersions_dialogOps_notifyOnError(  szTxt,  wfxVersions_dialogOps_localeParams( "szHException" ).replace( "[E]", eHTTP.name )  );

  document.getElementById(  "wfxVersions_dialogIDs_editHome"  ).value    = xOld.szEditHome;
  document.getElementById(  "wfxVersions_dialogIDs_editDown"  ).value    = xOld.szEditDown;
  document.getElementById(  "wfxVersions_dialogIDs_editURL"  ).value     = xOld.szEditURL;
  wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].home = xOld.szScanHome;
  wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].down = xOld.szScanDown;
  wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].url  = xOld.szScanURL;

  var  qSize = 0;
  var  cType = {  value:  ""  };
  try
  {
    try
    {
      var  xType = wfxVersions_dialogVars_httpRequest.getResponseHeader(  "Content-Type"  )
                                                     .replace(  /(\s+)/g,  ""  ).match(  /charset=([^;]+)(;|$)/i  );
      cType.value = xType[1];
    }
    catch(  e  )
    {}

    qSize = wfxVersions_globalOps_ConvertToUnicode(  wfxVersions_dialogVars_httpRequest.responseText,  cType  ).length;
  }
  catch(  e  )
  {}

  var  szOps = "Edit";
  if(  document.getElementById( "wfxVersions_dialogIDs_tabs" ).childNodes.length > 2  )
  {
    szOps = "Scan";
  }
  wfxVersions_fileOps_log(  szOps,
                            wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].list,
                            wfxVersions_dialogOps_lineFromItem( wfxVersions_dialogVars_progress ),
                            "",
                            szTxt,
                            qSize,
                            cType.value  );
}


/****************************************************************************/
/*! \brief  Prepares a HTTP Request from .list URLs
**
**  \param  --
**
**  \return  --
*/
function  wfxVersions_httpOps_headLaunch()
{
  if(  wfxVersions_dialogVars_itemsList == null  )
  {
    return;
  }
  if(  wfxVersions_dialogVars_httpRequest != null  )
  {
    wfxVersions_httpOps_ghost();
  }

  var  szURL2Scan = "";
  if(  document.getElementById( "wfxVersions_dialogIDs_tabs" ).childNodes.length > 2  )
  {
    for(  wfxVersions_dialogVars_progress = 0;  wfxVersions_dialogVars_progress < wfxVersions_dialogVars_itemsList.length;  wfxVersions_dialogVars_progress++  )
    {
      if(  wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].status == 0  )
      {
        break;
      }
    }

    if(  wfxVersions_dialogVars_progress >= wfxVersions_dialogVars_itemsList.length  )
    {
      wfxVersions_dialogOps_throbberIdle();
      return;
    }

    szURL2Scan = wfxVersions_dialogOps_likeMRT(  wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].url,  wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].version  );
    for(  var  i = 0;  i < wfxVersions_dialogVars_itemsList.length;  i++  )
    {
      var  szURL = wfxVersions_dialogOps_likeMRT(  wfxVersions_dialogVars_itemsList[i].url,  wfxVersions_dialogVars_itemsList[i].version  );
      if((  szURL == szURL2Scan  )&&(  wfxVersions_dialogVars_itemsList[i].status == 0  ))
      {
        wfxVersions_dialogVars_itemsList[i].step = 0;
        wfxVersions_dialogOps_scanStatus(  i,  2  );
      }
    }
    wfxVersions_dialogOps_throbberTabs();
  }
  else
  {
    wfxVersions_dialogVars_itemsList[ wfxVersions_dialogVars_progress ].step = 0;
    wfxVersions_dialogOps_scanStatus(  wfxVersions_dialogVars_progress,  2  );
    szURL2Scan = wfxVersions_dialogOps_likeMRT(  document.getElementById( "wfxVersions_dialogIDs_editURL" ).value,  document.getElementById( "wfxVersions_dialogIDs_editVersion" ).value  );
  }

  wfxVersions_httpOps_realLaunch(  szURL2Scan,  true  );
}


/****************************************************************************/
/*! \brief  Actual launch of a HTTP Request
**
**  XMLHttpRequest() function is used, which provides raw content of URLs
**  (works even with images or other resources)
**
**  \param  szURL     the URL to be scanned
**  \param  bTimeout  start timeout or not (if already started)
**
**  \return  --
*/
function  wfxVersions_httpOps_realLaunch(  szURL,  bTimeout  )
{
  try
  {
    /*  first we request the HEAD to determine the page encoding  */
    wfxVersions_dialogVars_httpRequest = new XMLHttpRequest();
    wfxVersions_dialogVars_httpRequest.onerror    = wfxVersions_httpOps_loadError;
    wfxVersions_dialogVars_httpRequest.onload     = wfxVersions_httpOps_loadReady;
    wfxVersions_dialogVars_httpRequest.onprogress = wfxVersions_httpOps_loadProgress;
    wfxVersions_dialogVars_httpRequest.open(  "GET",  szURL,  true  );

    /*  http://forums.mozillazine.org/viewtopic.php?f=19&t=501254&p=5768645  */
    wfxVersions_dialogVars_httpRequest.channel.loadFlags |= Components.interfaces.nsIRequest.LOAD_BYPASS_CACHE;
    try
    {
      /*  the order of these seems to matter (Mozilla Bugs 183470, 327790)
      **  and the HTTP 1.0 pragma=no-cache seems to be needed also
      */
      wfxVersions_dialogVars_httpRequest.setRequestHeader(  "Cache-Control",  "max-age=0, no-cache, no-store, must-revalidate"  );  /*  pre-check=0, post-check=0  */
      wfxVersions_dialogVars_httpRequest.setRequestHeader(  "Pragma",         "no-cache"  );
      wfxVersions_dialogVars_httpRequest.setRequestHeader(  "Expires",        "Wed, 11 Jan 1984 05:00:00 GMT"  );
    }
    catch(  e  )
    {
      /*  exception for ftp:/ URL-s on FF3  */
    }

    wfxVersions_dialogVars_httpRequest.overrideMimeType(  "text/plain; charset=x-user-defined"  );
    wfxVersions_dialogVars_httpRequest.send(  null  );

    if(  bTimeout  )
    {
      /*  "timeout disabled" configuration is handled by AbortHTTP()  */
      wfxVersions_httpOps_abort(  wfxVersions_dialogVars_progress,  false  );
    }
  }
  catch(  e  )
  {
    wfxVersions_httpOps_onError(  e  );
  }
}


/****************************************************************************/
/*! \brief  notification on page-loading error
**
**  The error reason is given in the object's .status member
**
**  \param[in]  eReq  onerror event for the HTTP Request in progress
**
**  \return  --
*/
function  wfxVersions_httpOps_loadError(  eReq  )
{
  try
  {
    var  reStatus = wfxVersions_dialogVars_httpRequest.status;
    if((  reStatus == 0/*ftp*/  )||(  reStatus == 200/*http*/  ))
    {
      /*  can happen e.g. when using a firewall  */
      wfxVersions_httpOps_loadReady();
      return;
    }

    var  xOld =
    {
      szEditHome:  document.getElementById(  "wfxVersions_dialogIDs_editHome"  ).value,
      szEditDown:  document.getElementById(  "wfxVersions_dialogIDs_editDown"  ).value,
      szEditURL:   document.getElementById(  "wfxVersions_dialogIDs_editURL"  ).value,
      szVersion:   wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].version,
      szScanHome:  wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].home,
      szScanDown:  wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].down,
      szScanURL:   wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].url
    };

    document.getElementById(  "wfxVersions_dialogIDs_editHome"  ).value    = document.getElementById(  "wfxVersions_dialogIDs_editHome"  ).previousSibling.getAttribute(  "tooltiptext"  );
    document.getElementById(  "wfxVersions_dialogIDs_editDown"  ).value    = document.getElementById(  "wfxVersions_dialogIDs_editDown"  ).previousSibling.getAttribute(  "tooltiptext"  );
    document.getElementById(  "wfxVersions_dialogIDs_editURL"  ).value     = document.getElementById(  "wfxVersions_dialogIDs_editURL"   ).previousSibling.getAttribute(  "tooltiptext"  );
    wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].home = wfxVersions_dialogOps_likeMRT(  xOld.szScanHome,  xOld.szVersion  );
    wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].down = wfxVersions_dialogOps_likeMRT(  xOld.szScanDown,  xOld.szVersion  );
    wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].url  = wfxVersions_dialogOps_likeMRT(  xOld.szScanURL,   xOld.szVersion  );

    var  szError = reStatus + " (" + wfxVersions_dialogVars_httpRequest.statusText + ")";
    var  szTxt = wfxVersions_dialogOps_localeParams(  "szErrorHTTP"  ).replace(  "[E]",  szError  );
    wfxVersions_dialogOps_notifyOnError(  szTxt,  wfxVersions_dialogOps_localeParams( "szHServer" ).replace( "[E]", szError )  );

    document.getElementById(  "wfxVersions_dialogIDs_editHome"  ).value    = xOld.szEditHome;
    document.getElementById(  "wfxVersions_dialogIDs_editDown"  ).value    = xOld.szEditDown;
    document.getElementById(  "wfxVersions_dialogIDs_editURL"  ).value     = xOld.szEditURL;
    wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].home = xOld.szScanHome;
    wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].down = xOld.szScanDown;
    wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].url  = xOld.szScanURL;

    var  qSize = 0;
    var  cType = {  value:  ""  };
    try
    {
      try
      {
        var  xType = wfxVersions_dialogVars_httpRequest.getResponseHeader(  "Content-Type"  )
                                                       .replace(  /(\s+)/g,  ""  ).match(  /charset=([^;]+)(;|$)/i  );
        cType.value = xType[1];
      }
      catch(  e  )
      {}

      qSize = wfxVersions_globalOps_ConvertToUnicode(  wfxVersions_dialogVars_httpRequest.responseText,  cType  ).length;
    }
    catch(  e  )
    {}

    var  szOps = "Edit";
    if(  document.getElementById( "wfxVersions_dialogIDs_tabs" ).childNodes.length > 2  )
    {
      szOps = "Scan";
    }
    wfxVersions_fileOps_log(  szOps,
                              wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].list,
                              wfxVersions_dialogOps_lineFromItem( wfxVersions_dialogVars_progress ),
                              "",
                              szTxt,
                              qSize,
                              cType.value  );
  }
  catch(  e  )
  {
    wfxVersions_httpOps_onError(  e  );
  }
}


/****************************************************************************/
/*! \brief  notification on page-loading progress
**
**  Sets all progress meters with equal URL to current page-loading status
**
**  \param[in]  e  the progress event
**
**  \return  --
*/
function  wfxVersions_httpOps_loadProgress(  e  )
{
  /*  abort scanning on close  */
  if((  wfxVersions_dialogVars_httpRequest == null  )||(  wfxVersions_dialogVars_itemsList == null  )||(  wfxVersions_dialogVars_itemsList.length == 0  ))
  {
    return;
  }

  /*  for pages of unknown size, totalSize = -1, which won't display nice  */
  var  iLen = 0x10000;
  if((  "totalSize" in e  )&&(  typeof(e.totalSize) == "number"  )&&(  e.totalSize < 0x7FFFFFFF  )&&(  e.totalSize > 0  ))
  {
    iLen = e.totalSize;
  }

  var  iPos = 0;
  if((  "position" in e  )&&(  typeof(e.position) == "number"  ))
  {
    iPos = e.position;
  }
  else
  {
    return;
  }

  if(  iPos > iLen  )
  {
    iPos = iLen;
  }
  var  iProgress = (  100 * iPos  ) / iLen;

  var  szMode = "determined";
  if(  iProgress < 1  )
  {
    szMode = "undetermined";
  }

  if(  document.getElementById( "wfxVersions_dialogIDs_tabs" ).childNodes.length > 2  )
  {
    for(  var  i = 0;  i < wfxVersions_dialogVars_itemsList.length;  i++  )
    {
      if((  wfxVersions_dialogVars_itemsList[i].url == wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].url  )
       &&(  wfxVersions_dialogVars_itemsList[i].status == 2  ))
      {
        document.getElementById(  "m" + i  ).setAttribute(  "value",  iProgress  );
        document.getElementById(  "m" + i  ).setAttribute(  "mode",   szMode  );
      }
    }
  }
  else
  {
    document.getElementById(  "wfxVersions_dialogIDs_editProgress"  ).setAttribute(  "value",  iProgress  );
    document.getElementById(  "wfxVersions_dialogIDs_editProgress"  ).setAttribute(  "mode",   szMode  );
  }
}


/****************************************************************************/
/*! \brief  Handles each application page as it's loaded
**
**  So basically for each application the following is done:
**  -# search for the start point inside page indicated by URL
**  -# from that point search for the "before version" pre-string
**  -# then search for the "after version" post-string
**  -# everything between pre ans post constitutes the version number/string
**  -# compare with old value (from .list file) and show if different (new)
**
**  \param  --
**
**  \return  --
*/
function  wfxVersions_httpOps_loadReady()
{
  /*  abort scanning on close  */
  if((  wfxVersions_dialogVars_httpRequest == null  )||(  wfxVersions_dialogVars_itemsList == null  )||(  wfxVersions_dialogVars_itemsList.length == 0  ))
  {
    return;
  }

  if((  wfxVersions_dialogVars_httpRequest.status != 200/*http*/  )
   &&(  wfxVersions_dialogVars_httpRequest.status !=   0/*ftp*/  ))
  {
    /*  can happen e.g. when using a proxy  */
    wfxVersions_httpOps_loadError(  null  );
    return;
  }

  /*  show a 100% progressmeter  */
  var  e = { totalSize: 1, position: 1 };
  wfxVersions_httpOps_loadProgress(  e  );

  var  szPage = "",  szURL = "";
  var  cType = {  value:  null  };
  try
  {
    try
    {
      var  xType = wfxVersions_dialogVars_httpRequest.getResponseHeader(  "Content-Type"  )
                                                     .replace(  /(\s+)/g,  ""  ).match(  /charset=([^;]+)(;|$)/i  );
      cType.value = xType[1];
    }
    catch(  e  )
    {}

    szPage = wfxVersions_globalOps_ConvertToUnicode(  wfxVersions_dialogVars_httpRequest.responseText,  cType  );
    szURL  = wfxVersions_dialogVars_httpRequest.channel.name;
  }
  catch(  e  )
  {
    wfxVersions_httpOps_onError(  e  );
    return;
  }

  /*  obey META Refresh with 0 seconds delay  */
  var  aMatch = szPage.match(  /<\s*META\s+HTTP-EQUIV\s*=\s*['"]?refresh['"]?\s+CONTENT\s*=\s*['"]?\s*0+\s*;\s*URL\s*=([^'"]+)['"]?\s*>/i  );
  if(  aMatch  )
  {
    /*  extract URL re-direction  */
    if(  aMatch[1].match( /.+\:[\/\\]{2}.+/ )  )
    {
      /*  absolute path given  */
      aMatch[0] = aMatch[1];
    }
    else
    {
      /*  relative path given  */
      var  iParent = szURL.lastIndexOf(  '\\'  );
      if(  iParent < 0  )
      {
        iParent = szURL.lastIndexOf(  '/'  );
      }
      aMatch[0] = szURL.substring(  0,  1 + iParent  ) + aMatch[1];
    }

    /*  URL clean-up  */
    aMatch[0] = aMatch[0].replace(  /([\/\\])\.[\/\\]/g,  "$1"  );
    while(  aMatch[0].match( /[\/\\]\.\.[\/\\]/ )  )
    {
      aMatch[1] = aMatch[0].replace(  /([\/\\])[^\/\\]+[\/\\]\.\.[\/\\]/g,  "$1"  );
      if(  aMatch[0] == aMatch[1]  )
      {
        break;
      }
      aMatch[0] = aMatch[1];
    }

    /*  new HTTP request with the new URL  */
    if(  szURL.toLowerCase() != aMatch[0].toLowerCase()  )
    {
      wfxVersions_globalOps_setTimeout(  "wfxVersions_httpOps_realLaunch('" + aMatch[0] + "',false)",  10  );
      return;
    }
  }

  /*  change to lower-case and escape all CR, LF or CR+LF  */
  szPage = szPage.replace(  /\r\n/g,  "&#10;"  ).replace(  /\r|\n/g,  "&#10;"  );
  var  szPageLower = szPage.toLowerCase();

  if(  document.getElementById( "wfxVersions_dialogIDs_tabs" ).childNodes.length > 2  )
  {
    /*  a single HTTP Request may correspond to multiple items
    **  this is detected by comparing URLs
    **  there's at least one item = current HTTP Request
    */
    szURL = wfxVersions_dialogOps_likeMRT(  wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].url,  wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].version  );
    for(  wfxVersions_dialogVars_progress = 0;  wfxVersions_dialogVars_progress < wfxVersions_dialogVars_itemsList.length;  wfxVersions_dialogVars_progress++  )
    {
      var  szLink = wfxVersions_dialogOps_likeMRT(  wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].url,  wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].version  );
      if((  szLink == szURL  )&&(  wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].status == 2  ))
      {
        var  oScan = wfxVersions_scanOps_itemSearch(  szPage,
                                                      wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].start,
                                                      wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].pre,
                                                      wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].post,
                                                      wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].dir % 2  );
        if(  oScan.status == 0  )
        {
          /*  simulate what Versions plugin does with the "version" string
          **  - remove all line-break characters
          **  - trim all spaces from beginning & end
          **  special treatment for the '|' character
          **  - replace by it's ASCII code
          */
          var  szUpdate = wfxVersions_globalOps_convertHTML(  oScan.result  );
          wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].update = szUpdate.replace(  /\|/g,  "&#124;"  );

          wfxVersions_dialogOps_scanStatus(  wfxVersions_dialogVars_progress,  5  );

          if(  szUpdate.toLowerCase() == wfxVersions_globalOps_convertHTML( wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].version ).toLowerCase()  )
          {
            /*  same version = no further action  */
            wfxVersions_dialogOps_hideAndFocus(  "x" + wfxVersions_dialogVars_progress,  "true"  );
          }
          else
          {
            /*  we need to auto-adjust the button-label width,
            **  because "crop" doesn't seem to work here
            */
            var  oItem = document.getElementById(  "x" + wfxVersions_dialogVars_progress  );
            var  szAppend = " ",  iLength = szUpdate.length,  iWidthComputed,  iWidthSet = oItem.parentNode.parentNode.style.getPropertyValue(  "width"  );
            oItem.parentNode.parentNode.style.setProperty(  "width",  "400px",  "important"  );
            if(  iLength > 100  )
            {
              iLength = 100;
              szAppend = unescape( "%u2026" ) + " ";
            }
            do
            {
              oItem.setAttribute(  "label",  " " + szUpdate.substring( 0, iLength ) + szAppend  );
              iLength--;
              szAppend = unescape( "%u2026" ) + " ";

              iWidthComputed = getComputedStyle(  oItem.parentNode.parentNode,  null  ).getPropertyValue(  "width"  );
            }
            while((  parseInt( iWidthComputed ) > 400  )&&(  iLength >= 0  ));
            oItem.parentNode.parentNode.style.setProperty(  "width",  iWidthSet,  "important"  );
            oItem.setAttribute(  "tooltiptext",  wfxVersions_dialogOps_localeParams( "szOverWrite" )  );

            /*  update TC MRT-style notations  */
            wfxVersions_dialogOps_updateScanLinks(  wfxVersions_dialogVars_progress,  wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].update  );
          }

          wfxVersions_fileOps_log(  "Scan",
                                    wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].list,
                                    wfxVersions_dialogOps_lineFromItem( wfxVersions_dialogVars_progress ),
                                    wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].update,
                                    "",
                                    szPage.length,
                                    cType.value  );
        }
        else
        {
          var  szHelp = "",  szTitle = "";
          switch(  oScan.status  )
          {
            /*  start-point not found means page content has changed
            **  which probably means there's a new "version" available
            */
            default:  /*  case  1:  */
              szHelp  = wfxVersions_dialogOps_localeParams(  "szHStart"  );
              szTitle = wfxVersions_dialogOps_localeParams(  "szStart"  );
            break;

            /*  before-version not found means page content has changed
            **  which probably means there's a new "version" available
            */
            case  2:
              szHelp  = wfxVersions_dialogOps_localeParams(  "szHBefore"  );
              szTitle = wfxVersions_dialogOps_localeParams(  "szPre"  );
            break;

            /*  after-version not found means page content has changed
            **  which probably means there's a new "version" available
            */
            case  3:
              szHelp  = wfxVersions_dialogOps_localeParams(  "szHAfter"  );
              szTitle = wfxVersions_dialogOps_localeParams(  "szPost"  );
            break;
          }

          wfxVersions_dialogOps_scanStatus(  wfxVersions_dialogVars_progress,  4  );
          document.getElementById(  "u" + wfxVersions_dialogVars_progress  ).value = szTitle;
          document.getElementById(  "h" + wfxVersions_dialogVars_progress  ).removeAttribute(  "hidden"  );
          document.getElementById(  "h" + wfxVersions_dialogVars_progress  ).setAttribute(  "oncommand",  "wfxVersions_globalOps_alert('" + szTitle + "','" + szHelp + "')"  );

          wfxVersions_fileOps_log(  "Scan",
                                    wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].list,
                                    wfxVersions_dialogOps_lineFromItem( wfxVersions_dialogVars_progress ),
                                    "",
                                    szTitle,
                                    szPage.length,
                                    cType.value  );
        }
      }
    }
    wfxVersions_dialogOps_throbberTabs();

    /*  launch next HTTP Request only after current request complete  */
    wfxVersions_globalOps_setTimeout(  "wfxVersions_httpOps_headLaunch()",  40  );
  }
  else
  {
    /*  Edit mode active  */
    wfxVersions_dialogOps_scanStatus(  wfxVersions_dialogVars_progress,  4  );

    var  oScan = wfxVersions_scanOps_itemSearch(  szPage,
                                                  document.getElementById( "wfxVersions_dialogIDs_editStart"    ).value,
                                                  document.getElementById( "wfxVersions_dialogIDs_editPre"      ).value,
                                                  document.getElementById( "wfxVersions_dialogIDs_editPost"     ).value,
                                                  document.getElementById( "wfxVersions_dialogIDs_editForwards" ).selectedIndex  );
    if(  oScan.status == 0  )
    {
      /*  simulate what Versions plugin does with the "version" string
      **  - remove all line-break characters
      **  - trim all spaces from beginning & end
      **  special treatment for the '|' character
      **  - replace by it's ASCII code
      */
      var  szUpdate = wfxVersions_globalOps_convertHTML(  oScan.result  );
      wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].update = szUpdate.replace(  /\|/g,  "&#124;"  );
      document.getElementById(  "wfxVersions_dialogIDs_editVersion"  ).value = szUpdate;
      wfxVersions_dialogOps_scanStatus(  wfxVersions_dialogVars_progress,  5  );
      wfxVersions_dialogOps_updateEditLinks();

      wfxVersions_fileOps_log(  "Edit",
                                wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].list,
                                wfxVersions_dialogOps_lineFromEdit( wfxVersions_dialogVars_progress ),
                                wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].update,
                                "",
                                szPage.length,
                                cType.value  );
    }
    else
    {
      var  szErrMsg = "";

      switch(  oScan.status  )
      {
        /*  start-point not found means page content has changed
        **  which probably means there's a new "version" available
        */
        default:  /*  case  1:  */
          szErrMsg = wfxVersions_dialogOps_localeParams(  "szStart"  );
          wfxVersions_dialogOps_notifyOnError(  szErrMsg,  wfxVersions_dialogOps_localeParams( "szHStart" )  );
        break;

        /*  before-version not found means page content has changed
        **  which probably means there's a new "version" available
        */
        case  2:
          szErrMsg = wfxVersions_dialogOps_localeParams(  "szPre"  );
          wfxVersions_dialogOps_notifyOnError(  szErrMsg,  wfxVersions_dialogOps_localeParams( "szHBefore" )  );
        break;

        /*  after-version not found means page content has changed
        **  which probably means there's a new "version" available
        */
        case  3:
          szErrMsg = wfxVersions_dialogOps_localeParams(  "szPost"  );
          wfxVersions_dialogOps_notifyOnError(  szErrMsg,  wfxVersions_dialogOps_localeParams( "szHAfter" )  );
        break;
      }

      wfxVersions_fileOps_log(  "Edit",
                                wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].list,
                                wfxVersions_dialogOps_lineFromEdit( wfxVersions_dialogVars_progress ),
                                "",
                                szErrMsg,
                                szPage.length,
                                cType.value  );
    }

    wfxVersions_dialogOps_throbberIdle();
    wfxVersions_dialogOps_setImageAttribute(  document.getElementById( "wfxVersions_dialogIDs_buttonEdit" ),  "chrome://wfx_versions/skin/wfx_versions_edit.gif"  );
  }
}


/****************************************************************************/
/*! \brief  Finds a valid TC plugin path from the window-box or the preference
**
**  Used by Browse button-event function
**
**  The path is searched also for (grand)parent directories if invalid:
**  - first try with actual pref value
**  - then with parent folder(s) of the pref value
**  - then try with textbox value
**  - then try with parent folder(s) of the textbox value
**
**  \param  --
**
**  \return  directory to be used for .displayDirectory in file-open dialog
*/
function  wfxVersions_fileOps_getValidPath()
{
  var  fsDir = Components.classes[ "@mozilla.org/file/local;1" ]
                         .createInstance(  Components.interfaces.nsILocalFile  );
  fsDir.QueryInterface(  Components.interfaces.nsIFile  );

  /*  try with textbox value  */
  var  szPath = document.getElementById(  "wfxVersions_dialogIDs_pathText"  ).value;

  /*  create a file-object for the Versions plugin folder  */
  var  fsDir = wfxVersions_fileOps_initWithPath(  szPath  );
  if(  fsDir  )
  {
    if((  fsDir.exists()  )&&(  fsDir.isDirectory()  ))
    {
      return  fsDir;
    }

    /*  try with parent directories of the textbox value  */
    do
    {
      try
      {
        fsDir = fsDir.parent;
      }
      catch(  e  )
      {
        bValidPath = false;  /*  we've reached volume root  */
        break;
      }
    }
    while(  !fsDir.exists()  );
  }

  if(  fsDir == null  )
  {
    /*  try with actual pref value  */
    szPath = Components.classes[ "@mozilla.org/preferences-service;1" ]
                       .getService(  Components.interfaces.nsIPrefBranch  )
                       .getComplexValue(  "extensions.wfx_Versions.tcPluginPath",  Components.interfaces.nsISupportsString  )
                       .data;
    fsDir = wfxVersions_fileOps_initWithPath(  szPath  );
    if(  fsDir == null  )
    {
      return  null;  /*  invalid path string, could not init()  */
    }

    if((  fsDir.exists()  )&&(  fsDir.isDirectory()  ))
    {
      return  fsDir;
    }

    /*  try with parent directories of the pref value  */
    do
    {
      try
      {
        fsDir = fsDir.parent;
      }
      catch(  e  )
      {
        return  null;  /*  we've reached volume root  */
      }
    }
    while(  !fsDir.exists()  );
  }

  return  fsDir;
}


/****************************************************************************/
/*! \brief  Opens an OS-specific "browse for path" dialog
**
**  Alternative method for configuring TC plugin path
**
**  \param  --
**
**  \return  --
*/
function  wfxVersions_fileOps_browsePath()
{
  /*  browse for folder with .list files  */
  var  fsPick = Components.classes[ "@mozilla.org/filepicker;1" ]
                          .createInstance( Components.interfaces.nsIFilePicker );
  fsPick.QueryInterface( Components.interfaces.nsIFilePicker );
  fsPick.init(  window,
                document.getElementById( "wfxVersions_dialogIDs_pathText" ).getAttribute( "tooltiptext" ),
                Components.interfaces.nsIFilePicker.modeGetFolder  );

  var  fsDir = wfxVersions_fileOps_getValidPath();
  if(  fsDir != null  )
  {
    fsPick.displayDirectory = fsDir;
  }

  /*  finally show the folder-pick dialog  */
  if(  fsPick.show() == Components.interfaces.nsIFilePicker.returnOK  )
  {
    document.getElementById(  "wfxVersions_dialogIDs_pathText"  ).value = fsPick.file.path;
    wfxVersions_dialogOps_loadPathsCombo(  2  );
  }
}


/****************************************************************************/
/*! \brief  Writes stuff into files
**
**  \param[in]  szList  the full path of the file written
**  \param[in]  szText  the contents of the file to be written
**
**  \return  written (true) or error e.g. read-only (false)
*/
function  wfxVersions_fileOps_storeList(  szList,  szText  )
{
  /*  create a file-object for the existing .list file  */
  var  fsList = wfxVersions_fileOps_initWithPath(  szList  );
  if(  fsList == null  )
  {
    wfxVersions_globalOps_alert(  wfxVersions_fileOps_leafName( szList ),  wfxVersions_dialogOps_localeParams( "szAccess" ) + "\n\n" + szList  );
    return  false;
  }

  if(  !fsList.exists()  )
  {
    /*  attempt to create the file if it does not exist  */
    try
    {
      /*  create a stream-object to start writing  */
      var  ioFile = Components.classes[ "@mozilla.org/network/file-output-stream;1" ]
                              .createInstance(  Components.interfaces.nsIFileOutputStream  );
      ioFile.init(  fsList,  0x02/*write*/ | 0x08/*create*/ | 0x10/*append*/,  0666,  0  );
      ioFile.close();
    }
    catch(  e  )
    {}
  }

  if(  !fsList.exists()  )
  {
    wfxVersions_globalOps_alert(  fsList.leafName,  wfxVersions_dialogOps_localeParams( "szNoExist" ) + "\n\n" + szList  );
    return  false;
  }

  if(  fsList.isDirectory()  )
  {
    wfxVersions_globalOps_alert(  fsList.leafName,  wfxVersions_dialogOps_localeParams( "szNoFile" ) + "\n\n" + szList  );
    return  false;
  }

  if(  !fsList.isWritable()  )
  {
    wfxVersions_globalOps_alert(  fsList.leafName,  wfxVersions_dialogOps_localeParams( "szFileWrite" ) + "\n\n" + szList  );
    return  false;
  }

  var  bDone = wfxVersions_fileOps_writeFile(  fsList,  szText  );
  if(  !bDone  )
  {
    wfxVersions_globalOps_alert(  fsList.leafName,  wfxVersions_dialogOps_localeParams( "szAccess" ) + "\n\n" + szList  );
  }
  return  bDone;
}


/****************************************************************************/
/*! \brief  Modifies the .list files with the new versions
**
**  Bit of a problem there: the list items are grouped by URL, and we can't
**  re-order them by 'element' member because the scanning might still be
**  in progress
**
**  \param[in]  iItem  the list item being updated
**
**  \return  --
*/
function  wfxVersions_fileOps_modifyList(  iItem  )
{
  var  bConfirmEnabled = Components.classes[ "@mozilla.org/preferences-service;1" ]
                                   .getService(  Components.interfaces.nsIPrefBranch  )
                                   .getBoolPref(  "extensions.wfx_Versions.confirm.scan.update"  );

  var  szConfirm = document.getElementById(  "x" + iItem  ).getAttribute(  "tooltiptext"  );
  if((  szConfirm.length != 0  )&&(  bConfirmEnabled  ))
  {
    if(  !wfxVersions_globalOps_confirm( wfxVersions_dialogVars_itemsList[iItem].line, szConfirm )  )
    {
      return;
    }
  }

  var  szVersion = wfxVersions_dialogVars_itemsList[iItem].version;
  wfxVersions_dialogVars_itemsList[iItem].version = wfxVersions_dialogVars_itemsList[iItem].update;

  /*  search all items from the same .list file  */
  var  txt = "";
  var  j;
  for(  j = 0;  j < wfxVersions_dialogVars_itemsList.length;  j++  )
  {
    if(  wfxVersions_dialogVars_itemsList[j].list == wfxVersions_dialogVars_itemsList[iItem].list  )
    {
      if(  txt.length != 0  )
      {
        txt += "\r\n";
      }
      txt += wfxVersions_dialogOps_lineFromItem(  j  );
    }
  }

  if(  wfxVersions_fileOps_storeList( wfxVersions_dialogVars_itemsList[iItem].list, txt )  )
  {
    wfxVersions_dialogOps_hideAndFocus(  "x" + iItem,  "true"  );

    /*  copy from 'new version' to 'old version'  */
    var  szVer = wfxVersions_globalOps_convertHTML(  wfxVersions_dialogVars_itemsList[iItem].version  );
    if(  szVer.length > 36  )
    {
      szVer = szVer.substring(  0,  36  ) + unescape( "%u2026" );
    }
    document.getElementById(  "y" + iItem  ).setAttribute(  "wfxIsError",  "false"  );
    document.getElementById(  "y" + iItem  ).value = szVer;

    /*  enlarge window to show entire content  */
    wfxVersions_dialogOps_resizeNow();
  }
  else
  {
    /*  file not written => restore version  */
    wfxVersions_dialogVars_itemsList[iItem].version = szVersion;
  }
}


/****************************************************************************/
/*! \brief  Create a new .list file
**
**  \param[in]  szOps  name of the operation performed, used in error alerts
**
**  \return  --
*/
function  wfxVersions_editOps_createFile(  szOps  )
{
  var  oInput = {  value:  wfxVersions_dialogOps_localeParams( "szNewList" )  };
  var  oCheck = {  value:  false  };

  var  oPrompts = Components.classes[ "@mozilla.org/embedcomp/prompt-service;1" ]
                            .getService(  Components.interfaces.nsIPromptService  );
  var  bResult = oPrompts.prompt(  window,  szOps,  wfxVersions_dialogOps_localeParams( "szToCreate" ),  oInput,  null,  oCheck  );

  if(  bResult  )
  {
    szList = oInput.value;
    if(  szList.match( /(\.list)$/g ) == null  )
    {
      szList += ".list";
    }

    try
    {
      wfxVersions_dialogVars_filesList[0].append(  szList  );
    }
    catch(  e  )
    {
      if(  wfxVersions_dialogVars_filesList[0].path.indexOf( "/" ) < 0  )
      {
        szList = wfxVersions_dialogVars_filesList[0].path + "\\" + szList;
      }
      else
      {
        szList = wfxVersions_dialogVars_filesList[0].path + "/" + szList;
      }
      wfxVersions_globalOps_alert(  szOps,  wfxVersions_dialogOps_localeParams( "szChars" ) + "\n\n" + szList  );
      return;
    }

    try
    {
      wfxVersions_dialogVars_filesList[0].create(  Components.interfaces.nsIFile.NORMAL_FILE_TYPE,  0777  );
    }
    catch(  e  )
    {
      wfxVersions_globalOps_alert(  szOps,  wfxVersions_dialogOps_localeParams( "szNoNewFile" ) + "\n\n" + wfxVersions_dialogVars_filesList[0].path  );
      return;
    }

    /*  edit this new file  */
    wfxVersions_editOps_start(  szOps  );
  }
}


/****************************************************************************/
/*! \brief  Delete a .list file
**
**  \param[in]  szOps  name of the operation performed, used in error alerts
**
**  \return  --
*/
function  wfxVersions_editOps_deleteFile(  szOps  )
{
  var  bConfirmEnabled = Components.classes[ "@mozilla.org/preferences-service;1" ]
                                   .getService(  Components.interfaces.nsIPrefBranch  )
                                   .getBoolPref(  "extensions.wfx_Versions.confirm.file.delete"  );

  var  szConfirm = wfxVersions_dialogOps_localeParams(  "szToDelete"  );
  if((  szConfirm.length != 0  )&&(  bConfirmEnabled  ))
  {
    szConfirm += "\n\n" + wfxVersions_dialogVars_filesList[0].path;
    if(  !wfxVersions_globalOps_confirm( szOps, szConfirm )  )
    {
      wfxVersions_memoOps_storeSelections(  "extensions.wfx_Versions.editSelections",  "extensions.wfx_Versions.editMemorize"  );
      return;
    }
  }

  var  iOldPerm = wfxVersions_dialogVars_filesList[0].permissions;
  try
  {
    wfxVersions_dialogVars_filesList[0].permissions |= 0200;  /*  make writable for owner  */
    wfxVersions_dialogVars_filesList[0].remove(  true  );
  }
  catch(  e  )
  {
    wfxVersions_globalOps_alert(  szOps,  wfxVersions_dialogOps_localeParams( "szNoDelete" ) + "\n\n" + wfxVersions_dialogVars_filesList[0].path  );
    try{  wfxVersions_dialogVars_filesList[0].permissions = iOldPerm;  }catch(e){}
    return;
  }

  var  oTabs = document.getElementById(  "wfxVersions_dialogIDs_tabs"  ).childNodes;
  if(  oTabs.length > 2  )
  {
    var  i;
    for(  i = 2;  i < oTabs.length;  i++  )
    {
      if(  oTabs[i].getAttribute( "tooltiptext" ) == wfxVersions_dialogVars_filesList[0].path  )
      {
        wfxVersions_dialogOps_closeTab(  wfxVersions_dialogVars_filesList[0].path  );
        return;
      }
    }
  }
}


/****************************************************************************/
/*! \brief  Test a string to determine if it's a valid path string
**
**  Some strings fail already at initWithPath(), others fail only at create()
**
**  \param[in]  szPath  the path-string to be tested
**
**  \return  string is valid/invalid, path exists/does not exist
**           - 0: string valid, path already exists
**           - 1: string valid, path does not exist
**           - 2: string invalid
*/
function  wfxVersions_fileOps_isValidPathString(  szPath  )
{
  /*  asterisk is used as splitter, so it's definitely not allowed  */
  if(  szPath.match( /\*/ )  )
  {
    return  2;
  }

  /*  disallow empty string(s)  */
  if((  szPath == ""  )||(  szPath == "\\"  )||(  szPath == "/"  ))
  {
    return  2;
  }

  var  fsPath = wfxVersions_fileOps_initWithPath(  szPath  );
  if(  fsPath == null  )
  {
    /*  could not create file object, definitely an invalid string  */
    return  2;
  }

  /*  path already exists, definitely a valid string  */
  if(  fsPath.exists()  )
  {
    return  0;
  }

  /*  from this point on, we can only determine that the string is valid if
  **  we attempt to create() the directory with this path, but harddisk I/O
  **  is overkill just for determining if a string can be a valid path
  */
  return  1;

  /*  attempt create directory from path  */
  try
  {
    var  fsParent = fsPath;
    while(  !fsParent.parent.exists()  )
    {
      fsParent = fsParent.parent;
    }

    fsPath.create(  Components.interfaces.nsIFile.DIRECTORY_TYPE,  0777  );

    /*  successfully created, so it's a valid string  */
    fsParent.remove(  true  );
    return  1;
  }
  catch(  e  )
  {
    if(  e.result != Components.results.NS_ERROR_FAILURE  )
    {
      /*  access denied, but assume valid string  */
      return  1;
    }
  }

  /*  failed to create directory due to unknown reason,
  **  assume invalid string
  */
  return  2;
}


/****************************************************************************/
/*! \brief  detect default new/empty path and copy mini .lists
**
**  Used at initial install, to copy the mini .lists from the install package
**
**  \param[in]  fsDir  \e nsIFile for the configured .lists path
**  \param[in]  szOps  name of the operation performed, used in error alerts
**  \param[in]  szErr  the error which caused this function to be called
**
**  \return  accessible nsILocalFile of the configured .lists directory
*/
function  wfxVersions_fileOps_createDefault(  fsDir,  szOps,  szErr  )
{
  var  bConfirmEnabled = Components.classes[ "@mozilla.org/preferences-service;1" ]
                                   .getService(  Components.interfaces.nsIPrefBranch  )
                                   .getBoolPref(  "extensions.wfx_Versions.confirm.create.default.lists"  );

  var  szConfirm = wfxVersions_dialogOps_localeParams(  "szNewPath"  );
  if((  szConfirm.length != 0  )&&(  bConfirmEnabled  ))
  {
    if(  !wfxVersions_globalOps_confirm( szOps, szErr + "\n\n" + szConfirm )  )
    {
      return  fsDir;
    }
  }

  /*  create the directory if it doesn't exist  */
  if(  !fsDir.exists()  )
  {
    try
    {
      fsDir.create(  Components.interfaces.nsIFile.DIRECTORY_TYPE,  0777  );
    }
    catch(  e  )
    {
      /*  access denied?  */
      wfxVersions_globalOps_alert(  szOps,  wfxVersions_dialogOps_localeParams( "szNoNewDir" ) + "\n\n" + fsDir.path  );
      return  null;
    }
  }

  /*  retrieve the mini .lists from the install package  */
  var  szNoWrite = "\n";
  var  aLists = [  "totalcmd.list",  "drivers.list",  "miscellaneous.list",
                   "mozilla.list",   "palm.list",     "others.list",
                   "online.list"  ];

  var  i;
  for(  i = 0;  i < aLists.length;  i++  )
  {
    /*  retrieve each list  */
    var  fsList = fsDir.clone();
    fsList.append(  aLists[i]  );
    try
    {
      fsList.create(  Components.interfaces.nsIFile.NORMAL_FILE_TYPE,  0777  );

      /*  copy the .list contents to corresponding new .list file  */
      var  ioReq = new XMLHttpRequest();
      ioReq.open(  "GET",  "chrome://wfx_versions/content/samples/" + aLists[i],  false  );
      ioReq.overrideMimeType(  "text/plain"  );
      ioReq.send(  null  );

      var  bDone = wfxVersions_fileOps_writeFile(  fsList,  ioReq.responseText  );
      if(  bDone  )
      {
        szNoWrite = null;
      }
      else
      {
        if(  szNoWrite  )
        {
          szNoWrite += "\n" + fsList.path;
        }
      }
    }
    catch(  e  )
    {
      if(  szNoWrite  )
      {
        szNoWrite += "\n" + fsList.path;
      }
    }
  }

  if(  szNoWrite  )
  {
    wfxVersions_globalOps_alert(  szOps,  wfxVersions_dialogOps_localeParams( "szNoNewFiles" ) + szNoWrite  );
  }
  else
  {
    return  wfxVersions_fileOps_openDir(  szOps  );
  }
  return  fsDir;
}


/****************************************************************************/
/*! \brief  writes a file with UTF-8 content
**
**  \param[in]  fsList  the file to be written
**  \param[in]  szText  the content to be written
**
**  \return  succeeded (true) / failed (false)
*/
function  wfxVersions_fileOps_writeFile(  fsList,  szText  )
{
  /*  create a stream-object to start writing  */
  var  ioFile = Components.classes[ "@mozilla.org/network/file-output-stream;1" ]
                          .createInstance(  Components.interfaces.nsIFileOutputStream  );
  ioFile.QueryInterface(  Components.interfaces.nsIOutputStream  );
  try
  {
    ioFile.init(  fsList,  0x02/*write-only*/ | 0x20/*truncate*/,  0664,  0  );
  }
  catch(  e  )
  {
    return  false;
  }

  var  ioStream = Components.classes[ "@mozilla.org/intl/converter-output-stream;1" ]
                            .createInstance(  Components.interfaces.nsIConverterOutputStream  );
  ioStream.init(  ioFile,  "UTF-8",  0,  "?".charCodeAt(0)  );

  /*  write all data without buffering  */
  ioStream.writeString(  szText  );

  ioStream.close();
  ioFile.close();
  return  true;
}


/****************************************************************************/
/*! \brief  searches for .list files inside the configured directory
**
**  This function:
**  - opens the configured directory, and issues corresponding alerts for
**    directory-access errors (the function returns \e null in case of errors)
**  - parses the directory contents and creates an array of files with names
**    ending in \e .list
**  - returns this array if not empty (i.e. if any .list files found)
**
**  The calling function must process the returned files e.g. to determine:
**  - if they are readable/writable (for scan/edit)
**  - if they have valid content (refer to .list file format description)
**
**  \param[in]  szOps  name of the operation performed, used in error alerts
**
**  \return  accessible nsILocalFile of the configured .lists directory
*/
function  wfxVersions_fileOps_openDir(  szOps  )
{
  while(  wfxVersions_dialogVars_filesList.length != 0  )
  {
    wfxVersions_dialogVars_filesList.pop();
  }

  /*  retrieve the configured path  */
  var  szPath = document.getElementById(  "wfxVersions_dialogIDs_pathText"  ).value;

  /*  create a file-object for the Versions plugin folder  */
  var  fsDir = wfxVersions_fileOps_initWithPath(  szPath  );
  if(  fsDir == null  )
  {
    /*  invalid path string  */
    wfxVersions_globalOps_alert(  szOps,  wfxVersions_dialogOps_localeParams( "szChars" ) + "\n\n" + szPath  );
    return  null;
  }

  if(  !fsDir.exists()  )
  {
    /*  path does not exist  */
    return  wfxVersions_fileOps_createDefault(  fsDir,  szOps,  wfxVersions_dialogOps_localeParams( "szDirNoExist" ) + "\n\n" + szPath  );
  }

  if(  !fsDir.isDirectory()  )
  {
    /*  path is not a directory  */
    wfxVersions_globalOps_alert(  szOps,  wfxVersions_dialogOps_localeParams( "szNoDir" ) + "\n\n" + szPath  );
    return  null;
  }

  var  fsEntries;
  try
  {
    fsEntries = fsDir.directoryEntries;
  }
  catch(  e  )
  {
    /*  access denied on this path  */
    wfxVersions_globalOps_alert(  szOps,  wfxVersions_dialogOps_localeParams( "szDirAccess" ) + "\n\n" + szPath  );
    return  null;
  }

  while(  fsEntries.hasMoreElements()  )
  {
    /*  create a file-object for each file/folder in the plugin folder  */
    var  fsList = fsEntries.getNext();
    fsList.QueryInterface(  Components.interfaces.nsIFile  );

    if((  fsList.isFile()  )&&(  fsList.path.match( /(\.list)$/g ) != null  ))
    {
      wfxVersions_dialogVars_filesList.push(  fsList  );
    }
  }

  if(  wfxVersions_dialogVars_filesList.length == 0  )
  {
    /*  no .list files found at the configured path  */
    return  wfxVersions_fileOps_createDefault(  fsDir,  szOps,  wfxVersions_dialogOps_localeParams( "szFile" ) + "\n\n" + szPath  );
  }

  return  fsDir;
}


/****************************************************************************/
/*! \brief  Copy currently edited item to another .list
**
**  \param[in]  szOps  name of the operation performed, used in error alerts
**
**  \return  --
*/
function  wfxVersions_editOps_copyStart(  szOps  )
{
  if(  wfxVersions_dialogVars_filesList.length == 0  )
  {
    return;
  }
  wfxVersions_memoOps_storeSelections(  "extensions.wfx_Versions.copySelections",  "extensions.wfx_Versions.copyMemorize"  );

  /*  create a file-object for the .list file opened  */
  var  fsList = wfxVersions_dialogVars_filesList[0];
  var  szPath = wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].list;

  /*  ask the user for confirmation (0-length string = don't ask)  */
  var  bConfirmEnabled = Components.classes[ "@mozilla.org/preferences-service;1" ]
                                   .getService(  Components.interfaces.nsIPrefBranch  )
                                   .getBoolPref(  "extensions.wfx_Versions.confirm.edit.copy"  );

  wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].list = fsList.path;
  var  szConfirm   = wfxVersions_dialogOps_localeParams(  "szAppend"  );
  var  szCopyOkMsg = wfxVersions_dialogOps_localeParams(  "szCopyOk"  );
  wfxVersions_dialogVars_itemsList[wfxVersions_dialogVars_progress].list = szPath;

  if((  szConfirm.length != 0  )&&(  bConfirmEnabled  ))
  {
    if(  !wfxVersions_globalOps_confirm( document.getElementById( "wfxVersions_dialogIDs_editCopy" ).getAttribute("tooltiptext"), szConfirm )  )
    {
      return;
    }
  }

  if(  !fsList.exists()  )
  {
    /*  attempt to create the file if it does not exist  */
    try
    {
      /*  create a stream-object to start writing  */
      var  ioFile = Components.classes[ "@mozilla.org/network/file-output-stream;1" ]
                              .createInstance(  Components.interfaces.nsIFileOutputStream  );
      ioFile.init(  fsList,  0x02/*write*/ | 0x08/*create*/ | 0x10/*append*/,  0666,  0  );
      ioFile.close();
    }
    catch(  e  )
    {}
  }

  if(  !fsList.exists()  )
  {
    wfxVersions_globalOps_alert(  fsList.leafName,  wfxVersions_dialogOps_localeParams( "szNoExist" ) + "\n\n" + fsList.leafName  );
    return;
  }

  if(  fsList.isDirectory()  )
  {
    wfxVersions_globalOps_alert(  fsList.leafName,  wfxVersions_dialogOps_localeParams( "szNoFile" ) + "\n\n" + fsList.leafName  );
    return;
  }

  if(  !fsList.isWritable()  )
  {
    wfxVersions_globalOps_alert(  fsList.leafName,  wfxVersions_dialogOps_localeParams( "szFileWrite" ) + "\n\n" + fsList.leafName  );
    return;
  }

  /*  create a stream-object for UTF-8 encoding  */
  var  ioFile = Components.classes[ "@mozilla.org/network/file-input-stream;1" ]
                          .createInstance(  Components.interfaces.nsIFileInputStream  );
  ioFile.QueryInterface(  Components.interfaces.nsILineInputStream  );
  ioFile.init(  fsList,  0x01/*read-only*/,  0444,  0  );

  var  ioStream = Components.classes[ "@mozilla.org/intl/converter-input-stream;1" ]
                            .createInstance(  Components.interfaces.nsIConverterInputStream  );
  ioStream.QueryInterface(  Components.interfaces.nsIUnicharLineInputStream  );
  ioStream.init(  ioFile,  "UTF-8",  1024,  0/*force exception on readLine encoding error*/  );

  /*  detect encoding  */
  var  bMoreLines;
  var  ozLine = {};
  try
  {
    do
    {
      bMoreLines = ioStream.readLine(  ozLine  );
    }
    while(  bMoreLines  );

    ioStream.close();
    ioFile.close();

    ioFile.init(  fsList,  0x01/*read-only*/,  0444,  0  );
    ioStream.init(  ioFile,  "UTF-8",  1024,  0  );
  }
  catch(  e  )
  {
    ioStream.close();
    ioFile.close();

    ioStream = ioFile;
    ioStream.init(  fsList,  0x01/*read-only*/,  0444,  0  );
  }

  var  txt = "";
  do
  {
    bMoreLines = ioStream.readLine(  ozLine  );
    if((  !bMoreLines  )&&(  ozLine.value.length == 0  ))
    {
      /*  no EOL (re-use the empty line from EOF)  */
    }
    else
    if(  txt.length != 0  )
    {
      txt += "\r\n";
    }
    else
    {
      /*  no EOL (first line in file)  */
    }
    txt += ozLine.value;
  }
  while(  bMoreLines  );
  ioStream.close();
  ioFile.close();

  /*  append the currently edited item to the text  */
  if(  txt.length != 0  )
  {
    txt += "\r\n";
  }
  txt += wfxVersions_dialogOps_lineFromItem(  wfxVersions_dialogVars_progress  );

  /*  finally...  */
  if(  wfxVersions_fileOps_writeFile( fsList, txt )  )
  {
    if(  fsList.path != szPath  )
    {
      /*  feedback message if copied to different .list  */
      if(  szCopyOkMsg.length != 0  )
      {
        wfxVersions_globalOps_alert(  szOps,  szCopyOkMsg  );
      }
      wfxVersions_editOps_clickEntry(  wfxVersions_dialogVars_progress  );
    }
    else
    {
      /*  refresh Edit tab if append in current .list  */
      txt = wfxVersions_dialogOps_lineFromItem(  wfxVersions_dialogVars_progress  );
      var  xItem;
      if(  wfxVersions_dialogOps_isComment( txt )  )
      {
        xItem = new wfxVersions_dialogOps_createNonItem(  txt,  szPath  );
      }
      else
      {
        xItem = new wfxVersions_dialogOps_createItem(  txt,  szPath  );
      }
      xItem.editrow = document.getElementById(  "wfxVersions_dialogIDs_editItems"  ).childNodes.length - 1;

      /*  copy the item with all properties  */
      wfxVersions_dialogVars_itemsList.pop();  /*  old new-line entry  */
      wfxVersions_dialogVars_itemsList.push(  xItem  );

      /*  append to Edit tab  */
      var  oName = document.getElementById(  "wfxVersions_dialogIDs_editItems"  ).lastChild;
      var  bHide = true;
      if(  oName.getAttribute( "hidden" ) != "true"  )
      {
        bHide = false;
      }
      else
      {
        oName.removeAttribute(  "hidden"  );
      }

      oName = oName.firstChild;
      oName.setAttribute(  "tooltiptext",  txt  );
      txt = wfxVersions_dialogVars_itemsList[ wfxVersions_dialogVars_progress ].line;
      if(  txt.length > 44  )
      {
        txt = txt.substring(  0,  44  ) + unescape( "%u2026" );
      }
      oName.value = txt;

      /*  we need another new-line entry  */
      wfxVersions_dialogOps_addEmpty();
      wfxVersions_editOps_clickEntry(  wfxVersions_dialogVars_itemsList.length - 2  );
      if(  bHide  )
      {
        document.getElementById(  "wfxVersions_dialogIDs_editItems"  ).lastChild.setAttribute(  "hidden",  "true"  );
      }
      wfxVersions_dialogOps_resizeNow();
      wfxVersions_editOps_sortList(  document.getElementById( "wfxVersions_dialogIDs_sortAuto" ).checked  );
    }
  }
  else
  {
    wfxVersions_globalOps_alert(  fsList.leafName,  wfxVersions_dialogOps_localeParams( "szAccess" ) + "\n\n" + fsList.leafName  );
  }
}


/****************************************************************************/
/*! \brief  Initialize a nsILocalFile with a given path
**
**  Try both *nix and win* style paths
**
**  Try both absolute and relative paths
**
**  \param[in]  szPath  the path to be attempted for init
**
**  \return  a \e nsILocalFile object, or \e null if error
*/
function  wfxVersions_fileOps_initWithPath(  szPath  )
{
  var  fsFile = Components.classes[ "@mozilla.org/file/local;1" ]
                          .createInstance(  Components.interfaces.nsILocalFile  );
  fsFile.QueryInterface(  Components.interfaces.nsIFile  );

  try
  {
    try
    {
      /*  try as absolute unmodified path  */
      fsFile.initWithPath(  szPath  );
    }
    catch(  e  )
    {
      try
      {
        /*  force absolute *nix style path  */
        fsFile.initWithPath(  szPath.replace( /\\/g, "/" )  );
      }
      catch(  e  )
      {
        /*  force absolute win* style path  */
        fsFile.initWithPath(  szPath.replace( /\//g, "\\" )  );
      }
    }
  }
  catch(  e  )
  {
    /*  create path relative to the profile directory  */
    var  szProfile = Components.classes[ "@mozilla.org/file/directory_service;1" ]
                               .getService(  Components.interfaces.nsIProperties  )
                               .get(  "ProfD",  Components.interfaces.nsIFile  )
                               .path;
    var  szRel = szProfile + "/" + szPath;
    if(  szProfile.indexOf( "/" ) < 0  )
    {
      szRel = szProfile + "\\" + szPath;
    }

    try
    {
      /*  try as relative unmodified path  */
      fsFile.initWithPath(  szRel  );
    }
    catch(  e  )
    {
      try
      {
        /*  force relative *nix style path  */
        fsFile.initWithPath(  szRel.replace( /\\/g, "/" )  );
      }
      catch(  e  )
      {
        try
        {
          /*  force relative win* style path  */
          fsFile.initWithPath(  szRel.replace( /\//g, "\\" )  );
        }
        catch(  e  )
        {
          return  null;  /*  invalid path string, could not init()  */
        }
      }
    }
  }

  /*  get rid of relative path constructs  */
  try{  fsFile.normalize();  }catch(e){}

  return  fsFile;
}


/****************************************************************************/
/*! \brief  Extracts the leaf-name from a given full-path string
**
**  \param[in]  szPath  the full-path string
**
**  \return  the leaf-name string
*/
function  wfxVersions_fileOps_leafName(  szPath  )
{
  var  aSplit = szPath.replace(  /\\/g,  "/"  ).split(  "/"  );
  return  aSplit[ aSplit.length - 1 ];
}


/****************************************************************************/
/*! \brief  Extracts the leaf-name from a given full-path string
**
**  \param[in]  szMode  current operation mode (Scan or Edit)
**  \param[in]  szList  path of .list file currently scanned
**  \param[in]  szLine  line from .list file
**  \param[in]  szPass  text for "all ok" (e.g. version update string)
**  \param[in]  szFail  text for failures (e.g. short error string)
**  \param[in]  szSize  size of the response
**  \param[in]  szCode  the charset of the response
**
**  \return  --
*/
function  wfxVersions_fileOps_log(  szMode,  szList,  szLine,  szPass,  szFail,  szSize,  szCode  )
{
  try
  {
    var  oPrefs = Components.classes[ "@mozilla.org/preferences-service;1" ]
                            .getService(  Components.interfaces.nsIPrefBranch  );

    var  bLog = oPrefs.getBoolPref(  "extensions.wfx_Versions.enable.logging." + szMode  );
    if(  bLog  )
    {
      var  szOps = document.getElementById(  "wfxVersions_dialogIDs_button" + szMode  ).getAttribute(  "label"  );
      var  xNow = new Date();

      /*  prepare log-file name  */
      var  szReName = oPrefs.getComplexValue(  "extensions.wfx_Versions.logging.name",  Components.interfaces.nsISupportsString  )
                            .data.replace(  "[Y]",  wfxVersions_globalOps_YMD( xNow )  )
                                 .replace(  "[M]",  szOps  );
      szReName = szList.replace(  /\.list$/i,  wfxVersions_globalOps_CLR( szReName )  );

      var  fsLog = wfxVersions_fileOps_initWithPath(  szReName  );

      var  ioFile = Components.classes[ "@mozilla.org/network/file-output-stream;1" ]
                              .createInstance(  Components.interfaces.nsIFileOutputStream  );
      ioFile.QueryInterface(  Components.interfaces.nsIOutputStream  );
      ioFile.init(  fsLog,  0x02/*write*/ | 0x08/*create*/ | 0x10/*append*/,  0666,  0  );

      var  ioStream = Components.classes[ "@mozilla.org/intl/converter-output-stream;1" ]
                                .createInstance(  Components.interfaces.nsIConverterOutputStream  );
      ioStream.init(  ioFile,  "UTF-8",  0,  "?".charCodeAt(0)  );

      /*  append one line  */
      var  szLog = oPrefs.getComplexValue(  "extensions.wfx_Versions.logging.format",  Components.interfaces.nsISupportsString  )
                         .data.replace(  "[Y]",  xNow.toUTCString()  )
                              .replace(  "[M]",  szOps  )
                              .replace(  "[J]",  szLine  )
                              .replace(  "[X]",  szCode  )
                              .replace(  "[Q]",  szSize  )
                              .replace(  "[U]",  szPass  )
                              .replace(  "[E]",  szFail  );

      ioStream.writeString(  szLog + "\r\n"  );

      ioStream.close();
      ioFile.close();
    }
  }
  catch(  e  )
  {}
}
